#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>

#if ENABLE_NLS
#include <libintl.h>
#define _(String) gettext( String )
#else
#define _(String) String
#endif
#define N_(String) String
#if HAVE_SETLOCALE
#include <locale.h>
#endif

#include <getopt.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#if HAVE_SYS_DIRENT_H
#include <sys/dirent.h>
#endif
#if HAVE_PATHS_H
#include <paths.h>
#include <sys/fcntl.h>
#else
#define _PATH_RWHODIR "/var/spool/rwho"
#include <sys/file.h>
#endif
#include <protocols/rwhod.h>

#define HOSTINC		16

typedef struct _tHost {
	int		nrUsers;	/* calculated during recv	*/
	struct	whod	host;		/* in host order!!		*/
} tHost;

enum 	sort_method { sHost, sLoad, sUser, sUptime, sLast };
char	*sortoption[] = { "hostname", "loadavg", "user", "uptime" };

static	tHost			*Hosts = NULL;
static	int			nrhosts = 0;
static	int			maxhosts = 0;
static	int			show_help;
static	int			show_version;
static	enum	sort_method	sort_method;
static	int			sort_reverse;
static	int			show_all;
static	char			*progname;
static	int			width;

static void readhosts(void);
static void sorthosts(void);
static void showhosts(void);

static
void usage(int status)
{
	if (status != 0) {
		fprintf(stderr,_("Try `%s --help' for more information.\n"),
			progname);
	}
	else {
		printf(_("Usage: %s [OPTION]...\n"),
			progname);
		printf(_("Show host status of remote machines.\n"
			"\n"
			"  -a  --show-all       show all users, instead of the not-idle ones\n"
			"  -h  --sort=hostname  sort according to the hostname\n"
			"  -l  --sort=loadavg   sort according to the loadaverage\n"
			"  -t  --sort=uptime    sort according to the uptime\n"
			"  -u  --sort=user      sort according to the number of users\n"
			"  -r  --sort-reverse   reverse the sort order\n"
			"      --help           display this help and exit\n"
			"      --version        output version information and exit\n"
			"\n"
			"Users are assumed idle when they have been inactive for more than\n"
			"an hour. Remote hosts for which no data is received in the last 11\n"
			"minutes are considered down.\n"));
	}
	exit(status);
}

int main(int argc, char *argv[])
{
static struct option longopt[] = {
	{ "help", no_argument, &show_help, 1},
	{ "version", no_argument, &show_version, 1},
	{ "show-all", no_argument, &show_all, 1},
	{ "sort", required_argument, NULL, 1},
	{ "sort-reverse", no_argument, &sort_reverse, 1},
	{ NULL } };
	int	ch;

#if HAVE_SETLOCALE
	setlocale(LC_MESSAGES, "");
#endif
#if ENABLE_NLS
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	progname = strrchr(argv[0],'/');
	if (progname) progname++;
	else progname = argv[0];

	while ((ch=getopt_long(argc,argv,"ahltur",longopt,NULL)) != -1)
	 switch (ch) {
	  case   0 :	break;		/* long option */
	  case   1 :	for (ch=0; ch<sLast; ch++)
				if (strcmp(optarg,sortoption[ch])==0) {
					sort_method = (enum sort_method)ch;
					break;
				}
			if (ch==sLast) {
				fprintf(stderr,_("%s: invalid sort option -- %s\n"),
					progname, optarg);
				usage(1);
			}
			break;
	  case 'a' :	show_all++;
			break;
	  case 'r' :	sort_reverse++;
			break;
	  case 'h' :	sort_method = sHost;
			break;
	  case 'l' :	sort_method = sLoad;
			break;
	  case 't' :	sort_method = sUptime;
			break;
	  case 'u' :	sort_method = sUser;
			break;
	  default  :	usage(1);
	 }

	if (show_version) {
		printf("%s - %s\n", progname, PACKAGE_VERSION);
		exit(0);
	}

	if (show_help)
		usage(0);

	readhosts();
	sorthosts();
	showhosts();
	return 0;
}

static
void readhosts()
{
	DIR	*dir;
	struct	dirent	*dirp;
#if HAVE_GETDENTS
	int	len;
	char	dirs[1024];
#endif

	width = 8;
	if (chdir(_PATH_RWHODIR) != 0) {
		perror(_PATH_RWHODIR);
		exit(1);
	}
	dir = opendir(".");
	if (dir==NULL) {
		perror("opendir");
		exit(1);
	}
#if HAVE_GETDENTS
while ((len=getdents(dirfd(dir),(struct dirent *)dirs,sizeof(dirs)))>0)
	for ((char *)dirp=dirs; (char *)dirp!=dirs+len; ((char *)dirp) += dirp->d_reclen) {
#else
	while ((dirp=readdir(dir)) != NULL) {
#endif
		tHost	*hostp;
		int	fd;
		int	size;

		if (strcmp(dirp->d_name,".")==0 ||
		    strcmp(dirp->d_name,"..")==0)
			continue;

		fd = open(dirp->d_name,O_RDONLY);
		if (fd<0) {
			perror(dirp->d_name);
			continue;
		}

		if (nrhosts == maxhosts) {
			maxhosts += HOSTINC;
			Hosts = realloc(Hosts,maxhosts*sizeof(tHost));
			if (Hosts == NULL) {
				fprintf(stderr, _("%s: too many hosts\n"),progname);
				exit(2);
			}
		}
		hostp = &Hosts[nrhosts];
		size = read(fd,&hostp->host,sizeof(struct whod));
		close(fd);
		/* did we get enough basic data? */
		if (size<60)
			continue;
		nrhosts++;

		/* quickly count the number of users */
		for (fd=0; size>60; fd++, size-= sizeof(struct whoent))
			if (show_all || hostp->host.wd_we[fd].we_idle<3600)
				hostp->nrUsers++;

		if (strlen(hostp->host.wd_hostname)>width)
			width = strlen(hostp->host.wd_hostname);
	}
	closedir(dir);
}

static
int sorthost(const void *a, const void *b)
{
	int	res = 0; /* stupid GCC: it is _always_ defined in the switch!! */

	switch (sort_method) {
	 case sUser :
		res = ((tHost *)b)->nrUsers - ((tHost *)a)->nrUsers;
		break;
	 case sLoad :
		res = ((tHost *)b)->host.wd_loadav[0] - ((tHost *)a)->host.wd_loadav[0];
		break;
	 case sUptime :
		res = ((tHost *)a)->host.wd_boottime - ((tHost *)b)->host.wd_boottime;
		break;
	 case sHost :
		res = strcmp(((tHost *)a)->host.wd_hostname, ((tHost *)b)->host.wd_hostname);
		break;
	 case sLast :
		res = 0;
		break;
	}

	if (sort_reverse) return -res;
	return res;
}

static
void sorthosts()
{
	qsort(Hosts,nrhosts,sizeof(tHost),sorthost);
}

static
char *time2str(char *buf, time_t diff)
{
	if (diff>86400L)
		sprintf(buf,"%3ld+%02ld:%02ld",
			diff/86400,
			(diff%86400)/3600,
			(diff%86400)/60%60);
	else
		sprintf(buf,"%6ld:%02ld",
			(diff/3600)%60,
			(diff/60)%60);
	return buf;
}

static
void showhosts()
{
	int	i;
	time_t	now, uptime;
	char	szTime[10];

	time(&now);
	uptime = now - 11*60;
	for (i=0; i<nrhosts; i++) {
		tHost	*ph = &Hosts[i];
		time_t	diff = ph->host.wd_sendtime - ph->host.wd_recvtime;

		if (ph->host.wd_recvtime<uptime)
			printf(_("%*s down %s\n"),
				-width, ph->host.wd_hostname,
				time2str(szTime,now-ph->host.wd_recvtime));
		else {
			const char *user = ph->nrUsers==1 ? _("user, ") : _("users,");
			printf(_("%*.*s   up %s, %5d %s  load %d.%02d, %d.%02d, %d.%02d\n"),
				-width, width, ph->host.wd_hostname,
				time2str(szTime,now-ph->host.wd_boottime+diff),
				ph->nrUsers,
				user,
				ph->host.wd_loadav[0]/100, ph->host.wd_loadav[0]%100,
				ph->host.wd_loadav[1]/100, ph->host.wd_loadav[1]%100,
				ph->host.wd_loadav[2]/100, ph->host.wd_loadav[2]%100);
		}
	}
}
